#include<stdio.h>
#include<stdlib.h>
#include<math.h>
#include"SimpleImage.h"
#include"BMPIO.h"
#define PI 3.141592653589793238462643383279502884197

/* DoG画像を生成 */
Image* makeDoGImage(int radii,double sig,double K){
  int gsx = 2*radii+1;
  int gsy = gsx;
  double sig2 = 2.0*sig*sig;
  double sigUV1 = 1.0/sig2;
  double sigUV2 = 1.0/(sig2*K*K);
  double coeff1 = 1.0/(sig2*PI);
  double coeff2 = 1.0/(sig2*PI*K*K);
  
  Image *tmp_img = new Image(gsx,gsy);
  int i,j;
  for(i=-radii;i<=radii;i++){
    int v = i+radii; int i2 = i*i;
    for(j=-radii;j<=radii;j++){
      int u = j+radii;
      double dist = ((double)(i2+j*j));
      tmp_img->img[v][u] = coeff1*exp(-(dist*sigUV1))-coeff2*exp(-(dist*sigUV2));
      
    }
  }
  
  return tmp_img;
}
/* 畳み込み */
void Conv(Image *in,int radii,Image *kernel,Image *out){
  int i,j;
  for(i=0;i<in->sy;i++)
    for(j=0;j<in->sx;j++){
      int i2,j2;
      out->img[i][j] = 0.0;
      for(i2=-radii;i2<=radii;i2++){
	int v = i2+radii;
	int v2 = i+i2;if(v2<0)v2=0;if(v2>in->sy-1)v2=in->sy-1;
	for(j2=-radii;j2<=radii;j2++){
	    int u = j2+radii;
	    int u2 = j+j2;if(u2<0)u2=0;if(u2>in->sx-1)u2=in->sx-1;
	    out->img[i][j] += in->img[v2][u2]*kernel->img[v][u];
	    
	  }
      }
      
    }
}

/* g++ -o DoGEdge DoGEdge.cxx -lm
引数5個
./DoGEdge 入力ファイル名 出力ファイル名 畳み込みの半径(int) DoGの標準偏差(double) DoGのバンド幅(double)
例えば
./DoGEdge lena.bmp test_out.bmp 10 0.5 2.0

*/

int main(int argc,char *argv[]){
  int i,j;
  int sx,sy;
  /* BMP画像の読み込み */
  BMPIO *myio = new BMPIO();
  myio->readBMPSize(&sx,&sy,argv[1]);
  
   /* 出力画像 */
  Image *R = new Image(sx,sy);
  Image *G = new Image(sx,sy);
  Image *B = new Image(sx,sy);
  myio->readBMP(R,G,B,argv[1]);
   /* 計算の為の画像 */
  Image *gray = new Image(sx,sy);
  Image *out_gray = new Image(sx,sy);
  
   /* グレースケール画像の作成 */
  for(i=0;i<R->sy;i++)
    for(j=0;j<R->sx;j++)
      gray->img[i][j] = (R->img[i][j]+G->img[i][j]+B->img[i][j])/3.0;
  
  
  Image *outR = new Image(sx,sy);
  Image *outG = new Image(sx,sy);
  Image *outB = new Image(sx,sy);
  
  /* 畳み込みの半径 */
  int radii = atoi(argv[3]);
  /* DoGの空間標準偏差(double) */
  double sig = atof(argv[4]);
  /* DoGのバンド幅(double) */
  double K = atof(argv[5]);

  /* DoG画像の保存 */
  Image *dog = makeDoGImage(radii,sig,K);
  
  
  /* 畳み込み */
  Conv(gray,radii,dog,out_gray);
  
  
  for(i=0;i<out_gray->sy;i++)
    for(j=0;j<out_gray->sx;j++){
      if(out_gray->img[i][j]>0.0){
	outR->img[i][j] = 255.0;
	outG->img[i][j] = 255.0;
	outB->img[i][j] = 255.0;
	
      }else{
	/* Edge */

	outR->img[i][j] = 0.0;
	outG->img[i][j] = 0.0;
	outB->img[i][j] = 0.0;
	
      }
    }

  /* 画像の保存 */
  myio->saveBMP(outR,outG,outB,argv[2]);
  
  /* メモリの開放 */
  delete dog;
  delete gray;
  delete out_gray;
    
  delete R;
  delete G;
  delete B;

  delete outR;
  delete outG;
  delete outB;

  delete myio;
  return 0;
}

#undef PI
